import logging
from typing import List, Dict, Any, Optional
from pymilvus import (
    connections, Collection, CollectionSchema, FieldSchema,
    DataType, list_collections, utility
)

class MilvusClient:
    def __init__(self, host: str = "localhost", port: str = "19530", alias: str = "default"):
        try:
            connections.connect(alias=alias, host=host, port=port)
            self.alias = alias
            logging.info(f"Connected to Milvus at {host}:{port}")
        except Exception as e:
            logging.error(f"Failed to connect to Milvus: {e}")
            raise

    def get_collection(self, name: str) -> Optional[Collection]:
        if name not in list_collections(using=self.alias):
            logging.warning(f"Collection '{name}' does not exist.")
            return None
        return Collection(name, using=self.alias)

    def create_collection(self, name: str, dim: int = 64) -> None:
        if name in list_collections(using=self.alias):
            logging.info(f"Collection '{name}' already exists.")
            return

        schema = CollectionSchema(
            fields=[
                FieldSchema(name="id", dtype=DataType.INT64, is_primary=True, auto_id=True),
                FieldSchema(name="vector", dtype=DataType.FLOAT_VECTOR, dim=dim),
            ],
            description=f"Collection '{name}' with dim={dim}",
            enable_dynamic_field=True
        )
        Collection(name=name, schema=schema, using=self.alias)
        logging.info(f"Created collection '{name}' with dimension {dim}.")

    def create_index(self, name: str, index_type: str = "IVF_FLAT", nlist: int = 128) -> bool:
        collection = self.get_collection(name)
        if not collection:
            return False

        if collection.indexes:
            logging.info(f"Collection '{name}' already has an index.")
            return False

        try:
            index_params = {
                "index_type": index_type,
                "metric_type": "L2",
                "params": {"nlist": nlist},
            }
            collection.create_index(field_name="vector", index_params=index_params)
            logging.info(f"Created index for collection '{name}'.")
            return True
        except Exception as e:
            logging.error(f"Failed to create index: {e}")
            return False

    def insert_vectors(
        self,
        name: str,
        vectors: List[List[float]],
        extra_fields: Optional[List[Dict[str, Any]]] = None,
        auto_flush: bool = True,
        auto_load: bool = True
    ) -> Optional[List[int]]:
        collection = self.get_collection(name)
        if not collection:
            return None

        if extra_fields is None:
            extra_fields = [{} for _ in vectors]

        if len(vectors) != len(extra_fields):
            logging.error("Mismatch between number of vectors and extra_fields.")
            return None

        insert_data = [{"vector": v, **meta} for v, meta in zip(vectors, extra_fields)]

        try:
            res = collection.insert(insert_data)
            if auto_flush:
                collection.flush()
            if auto_load:
                collection.load()
            return list(res.primary_keys)
        except Exception as e:
            logging.error(f"Insert failed for collection '{name}': {e}")
            return None

    def search_vectors(
        self,
        name: str,
        query_vector: List[float],
        top_k: int = 5,
        output_fields: Optional[List[str]] = None,
        filter_expr: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        collection = self.get_collection(name)
        if not collection:
            return []

        try:
            collection.load()
            results = collection.search(
                data=[query_vector],
                anns_field="vector",
                param={"metric_type": "L2", "params": {"nprobe": 10}},
                limit=top_k,
                output_fields=output_fields,
                expr=filter_expr
            )

            hits = []
            for result in results[0]:
                record = {"id": result.id, "distance": result.distance}
                if output_fields:
                    record.update(result.entity.get(field, None) for field in output_fields)
                hits.append(record)
            return hits
        except Exception as e:
            logging.error(f"Search failed in '{name}': {e}")
            return []

    def count_entities(self, name: str) -> Optional[int]:
        collection = self.get_collection(name)
        if not collection:
            return None

        try:
            collection.load()
            return collection.num_entities
        except Exception as e:
            logging.error(f"Count failed for '{name}': {e}")
            return None

    def delete_collection(self, name: str) -> bool:
        if name not in list_collections(using=self.alias):
            logging.warning(f"Collection '{name}' does not exist.")
            return False

        try:
            Collection(name=name, using=self.alias).drop()
            logging.info(f"Deleted collection '{name}'.")
            return True
        except Exception as e:
            logging.error(f"Failed to delete collection '{name}': {e}")
            return False

    def list_collections(self) -> List[str]:
        try:
            return list_collections(using=self.alias)
        except Exception as e:
            logging.error(f"Failed to list collections: {e}")
            return []
    
    def get_collection_schema(self, name: str) -> Optional[dict]:
        try:
            collection = Collection(name)
            return {
                "fields": {
                    f.name: {
                        "type": f.dtype.name,
                        "is_primary": f.is_primary,
                        "auto_id": getattr(f, "auto_id", None)
                    }
                    for f in collection.schema.fields
                },
                "description": collection.schema.description
            }
        except Exception as e:
            logging.error(f"Failed to get schema for '{name}': {e}")
            return None


import pytest
import numpy as np
from milvus_helper import MilvusClient
from pymilvus import list_collections, Collection

@pytest.fixture(scope="module")
def client():
    c = MilvusClient()
    yield c
    # Cleanup all test collections after all tests
    for name in list_collections():
        if name.startswith("test_"):
            Collection(name).drop()

@pytest.fixture(scope="module")
def setup_collection(client):
    name = "test_client_collection"
    client.delete_collection(name)
    client.create_collection(name, dim=64)
    vectors = np.random.rand(3, 64).tolist()
    metadata = [{"source": "test"} for _ in range(3)]
    client.insert_vectors(name, vectors, metadata)
    client.create_index(name)
    return name

def test_create_collection(client):
    name = "test_create"
    client.delete_collection(name)
    client.create_collection(name, dim=32)
    assert name in client.list_collections()

def test_create_collection_duplicate(client):
    name = "test_create"
    client.create_collection(name, dim=32)  # duplicate creation should pass silently

def test_create_collection_invalid_dim(client):
    with pytest.raises(Exception):
        client.create_collection("invalid", dim=-1)

def test_insert_vectors_success(client, setup_collection):
    vectors = np.random.rand(2, 64).tolist()
    metadata = [{"source": "unit"} for _ in range(2)]
    result = client.insert_vectors(setup_collection, vectors, metadata)
    assert isinstance(result, list)
    assert len(result) == 2

def test_insert_empty_vectors(client, setup_collection):
    result = client.insert_vectors(setup_collection, [], [])
    assert result is None

def test_insert_vectors_mismatch(client, setup_collection):
    vectors = np.random.rand(2, 64).tolist()
    metadata = [{"only": "one"}]
    result = client.insert_vectors(setup_collection, vectors, metadata)
    assert result is None

def test_search_vectors_success(client, setup_collection):
    vec = np.random.rand(64).tolist()
    results = client.search_vectors(setup_collection, vec, top_k=2, output_fields=["source"])
    assert isinstance(results, list)
    if results:
        assert "source" in results[0]
        assert "distance" in results[0]

def test_search_wrong_dim(client, setup_collection):
    vec = np.random.rand(10).tolist()
    results = client.search_vectors(setup_collection, vec)
    assert results == []

def test_search_nonexistent_collection(client):
    vec = np.random.rand(64).tolist()
    results = client.search_vectors("no_collection", vec)
    assert results == []

def test_get_collection_schema(client, setup_collection):
    schema = client.get_collection_schema(setup_collection)
    assert isinstance(schema, dict)
    assert "fields" in schema

def test_get_collection_schema_nonexistent(client):
    assert client.get_collection_schema("fake_collection") is None

def test_count_entities(client, setup_collection):
    count = client.count_entities(setup_collection)
    assert isinstance(count, int)
    assert count >= 3

def test_delete_collection(client):
    name = "test_delete"
    client.create_collection(name, dim=64)
    result = client.delete_collection(name)
    assert result
    assert name not in client.list_collections()

def test_delete_collection_nonexistent(client):
    result = client.delete_collection("no_such_collection")
    assert result is False

def test_get_collection(client, setup_collection):
    collection = client.get_collection(setup_collection)
    assert isinstance(collection, Collection)

def test_get_collection_nonexistent(client):
    assert client.get_collection("ghost") is None

def test_list_collections(client, setup_collection):
    collections = client.list_collections()
    assert isinstance(collections, list)
    assert setup_collection in collections
